// nested.hpp

// Implements a graph that supports ports and nesting.


//////////////////////////////////////////////////////////////
//
//  Copyright 2000   Simple Investments. 
//
//  Released into the public domain 27th June 2002.
//  Please change, sell, claim as your own, whatever you like.
//
//////////////////////////////////////////////////////////////

/*! private constructor for nested_graph, used when a subgraph is created to pass the parent graph details
    to the child.
*/
template <NESTED_TARGS> 
nested_graph<NESTED_ARGS>::nested_graph<NESTED_ARGS>(nested_graph<NESTED_ARGS> &pParent)
{
  // use this port list rather than creating a new one
  portList = pParent.portList;
  nestLevel = pParent.nestLevel + 1;
  parentGraph = &pParent;
}

/*! default constructor for nested_graph
*/
template <NESTED_TARGS> 
nested_graph<NESTED_ARGS>::nested_graph<NESTED_ARGS>()
{
  nestLevel = 0;
  portList = new PortGraph();
  parentGraph = NULL;
}

/*! default destructor for nested_graph, will destroy any remaining vertices in the graph, 
    to ensure no unecessary subgraphs,ports or edges remain.
*/
template <NESTED_TARGS> 
nested_graph<NESTED_ARGS>::~nested_graph<NESTED_ARGS>()
{
  // should probably delete all the vertices too, to clean up subgraphs.
  vertex_iterator i, iend;
  boost::tie(i, iend) = vertices(*this);
  while (i != iend) {
    remove_vertex(*this, *i);
    boost::tie(i, iend) = vertices(*this);
  }

  // if this is a top level graph then delete the port list.
  if (nestLevel == 0) {
    delete portList;
    portList = NULL;
  }
  
}

//! adds an atomic vertex to the graph
template <NESTED_TARGS> 
  nested_graph<NESTED_ARGS>::vertex_descriptor  add_vertex(nested_graph<NESTED_ARGS> &g, const VertexPlugin &x = VertexPlugin())
{
  typedef nested_graph<NESTED_ARGS> NG;
  NG::BaseGraph& base = *((NG::BaseGraph*)&g);

  return ggcl::add_vertex(base, NG::FinalVertexPlugin(NG::PortList(), NG::GraphPlugin(NULL, x)));
}

//! adds a composite vertex, ie a vertex with a sub graph
template <NESTED_TARGS> 
  nested_graph<NESTED_ARGS>::vertex_descriptor  add_composite_vertex(nested_graph<NESTED_ARGS> &g, const VertexPlugin &x = VertexPlugin())
{
  typedef nested_graph<NESTED_ARGS> NG;
  NG::BaseGraph& base = *((NG::BaseGraph*)&g);

  // create a new graph to be added to the vertex
  NG *g2 = new NG(g);

  // add the vertex with a pointer to the new graph
  NG::vertex_descriptor v = ggcl::add_vertex(base, NG::FinalVertexPlugin(NG::PortList(), NG::GraphPlugin(g2, x)));

  g2->parentVertex = v;

  return v;
}
 

//! retrieves the sub graph form a composite vertex
template <NESTED_TARGS> 
  nested_graph<NESTED_ARGS> *get_child_graph(nested_graph<NESTED_ARGS> &g, 
                                                                   nested_graph<NESTED_ARGS>::vertex_descriptor v)
{
  // get the property accessor for graph tag
  vertex_property_accessor<nested_graph<NESTED_ARGS>, vert_subgraph_tag>::type subGraph
    = get_vertex_property(g, vert_subgraph_tag());

  return subGraph[v];
}

//! returns true if the vertex has a subgraph
template <NESTED_TARGS> 
  bool is_composite_vertex(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS>::vertex_descriptor v)
{
  return get_child_graph(g, v) != NULL;
}

//! adds a port to an atomic vertex.
template <NESTED_TARGS> 
  nested_graph<NESTED_ARGS>::port_descriptor  add_port(nested_graph<NESTED_ARGS> &g, 
                                                                             nested_graph<NESTED_ARGS>::vertex_descriptor v,
                                                                             const PortPlugin &PP = PortPlugin())
{
  typedef nested_graph<NESTED_ARGS> NG;
  NG::BaseGraph& base = *((NG::BaseGraph*)&g);

  NG::edge_descriptor e;    // empty edge descriptor.

  // add a vertex to the portList object which will represent the port.
  NG::port_descriptor p = add_vertex(*g.portList, NG::FinalPortPlugin(&g, NG::PortEdgeGraphPlugin(NULL, NG::PortEdgePlugin(e, NG::PortVertexPlugin(v, PP)))));

  // add the port to the vertex's list of ports.
  // get the property accessor for port list tag
  vertex_property_accessor<nested_graph<NESTED_ARGS>, vert_portlist_tag>::type plist
    = get_vertex_property(g, vert_portlist_tag());

  plist[v].push_back(p);

  return p;
}

template <NESTED_TARGS> 
bool find_common_graph(nested_graph<NESTED_ARGS>* &g1, nested_graph<NESTED_ARGS>::vertex_descriptor &v1,
                       nested_graph<NESTED_ARGS>* &g2, nested_graph<NESTED_ARGS>::vertex_descriptor &v2)
{
  int nest1 = g1->nestLevel;
  int nest2 = g2->nestLevel;

  // now we have the vertices and ports these guys belong to, find the lowest common subgraph
  while (g1 != g2 && (nest1 != 0 || nest2 != 0))
  {
    // climb up the vertice with the lowest nesting level.
    if (nest1 > nest2) {
      v1 = g1->parentVertex;
      g1 = g1->parentGraph;
      nest1 = g1->nestLevel;
    } else {
      v2 = g2->parentVertex;
      g2 = g2->parentGraph;
      nest2 = g2->nestLevel;
    }
  }

  return g1 == g2;
}

/*! adds an edge to a graph connecting two ports. an edge is added between the lowest two vertices in the same subgraph.
 returns the edge and the sub graph that contains the edge */
template <NESTED_TARGS> 
  pair<nested_graph<NESTED_ARGS>::edge_descriptor, nested_graph<NESTED_ARGS>*> 
                                                                          add_edge_from_ports(nested_graph<NESTED_ARGS> &g, 
                                                                                   nested_graph<NESTED_ARGS>::port_descriptor p1,
                                                                                   nested_graph<NESTED_ARGS>::port_descriptor p2,
                                                                                   const EdgePlugin &EP = EdgePlugin())
{
  typedef nested_graph<NESTED_ARGS> NG;

  // get all the 
  vertex_property_accessor<NG::PortGraph, port_vertex_tag>::type vlist
    = get_vertex_property(*g.portList, port_vertex_tag());
  vertex_property_accessor<NG::PortGraph, port_vgraph_tag>::type glist
    = get_vertex_property(*g.portList, port_vgraph_tag());
  vertex_property_accessor<NG::PortGraph, port_edge_tag>::type elist
    = get_vertex_property(*g.portList, port_edge_tag());
  vertex_property_accessor<NG::PortGraph, port_egraph_tag>::type eglist
    = get_vertex_property(*g.portList, port_egraph_tag());

  // find the two vertices that will be joined
  NG::vertex_descriptor v1 = vlist[p1];
  NG  *g1 = glist[p1];
  NG::vertex_descriptor v2 = vlist[p2];
  NG  *g2 = glist[p2];

  // if either port is connected to an edge then it's very bad
    assert(elist[p1] == NG::edge_descriptor() && elist[p2] == NG::edge_descriptor());

  // find the lowest common subgraph.
  find_common_graph(g1, v1, g2, v2);

  // if we got to the top and they're still not equal then there's something 
  // seriously wrong with the graph
  assert(g1 == g2);
    
  // add the new edge
  EdgePlugin x;
  NG::FinalEdgePlugin ppp(NG::PortPair(p1, p2), EP);
  NG::edge_descriptor e;
  bool bOK;
  NG::BaseGraph& base = *((NG::BaseGraph*)g1);
  boost::tie(e, bOK) = add_edge(base, v1, v2, ppp);

  // now we know the edge we can add the info to the ports.
  elist[p1] = e;
  elist[p2] = e;
  eglist[p1] = g1;
  eglist[p2] = g2;

  return make_pair(e, g1);
}

/*!  adds an edge between two vertices (ports will be automatically added to the two vertices.
     returns the edge and the sub graph that contains the edge
*/
template <NESTED_TARGS> 
  pair<nested_graph<NESTED_ARGS>::edge_descriptor, nested_graph<NESTED_ARGS>*> 
                                                                          add_edge(nested_graph<NESTED_ARGS> &g, 
                                                                                   nested_graph<NESTED_ARGS>::vertex_descriptor v1,
                                                                                   nested_graph<NESTED_ARGS>::vertex_descriptor v2,
                                                                                   const EdgePlugin &EP = EdgePlugin())
{
  typedef nested_graph<NESTED_ARGS> NG;

  NG::port_descriptor p1 = add_port(g, v1);
  NG::port_descriptor p2 = add_port(g, v2);

  return add_edge_from_ports(g, p1, p2, EP);
}

/*! adds an edge between two vertices on different subgraphs(ports will be automatically added to the two vertices).
    returns the edge and the sub graph that contains the edge
*/
template <NESTED_TARGS> 
  pair<nested_graph<NESTED_ARGS>::edge_descriptor, nested_graph<NESTED_ARGS>*> 
                                                                          add_edge(nested_graph<NESTED_ARGS> &g1, 
                                                                                   nested_graph<NESTED_ARGS>::vertex_descriptor v1,
                                                                                   nested_graph<NESTED_ARGS> &g2, 
                                                                                   nested_graph<NESTED_ARGS>::vertex_descriptor v2,
                                                                                   const EdgePlugin &EP = EdgePlugin())
{
  typedef nested_graph<NESTED_ARGS> NG;

  NG::port_descriptor p1 = add_port(g1, v1);
  NG::port_descriptor p2 = add_port(g2, v2);

  return add_edge_from_ports(g1, p1, p2, EP);
}

template <NESTED_TARGS, class Tag> 
port_property_accessor<nested_graph<NESTED_ARGS>::PortGraph, Tag>::type get_port_property(nested_graph<NESTED_ARGS> &g, Tag t)
{
  return get_vertex_property(*g.portList, t);
}

// removes an edge from the graph, will leave the ports, but should disconnect them.
template <NESTED_TARGS> 
void remove_edge(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS>::edge_descriptor e)
{
  typedef nested_graph<NESTED_ARGS> NG;

  // first find the two ports from the edge and disconnect them from the edge, then remove the edge.
  NG::PortPair pp = ports(g, e);

  // remove edge from ports
  port_property_accessor<NG::PortGraph,port_edge_tag>::type edges
    = get_port_property(g, port_edge_tag());
  port_property_accessor<NG::PortGraph,port_egraph_tag>::type edgegraphs
    = get_port_property(g, port_egraph_tag());
  
  edges[pp.first] = NG::edge_descriptor();
  edges[pp.second] = NG::edge_descriptor();
  edgegraphs[pp.first] = NULL;
  edgegraphs[pp.second] = NULL;

  // now remove the actual edge and we're done
  NG::BaseGraph& base = *((NG::BaseGraph*)&g);
  remove_edge(base, e);
  
}

/*! removes the edge which connects the two ports */
template <NESTED_TARGS> 
void remove_edge(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS>::port_descriptor p1, nested_graph<NESTED_ARGS>::port_descriptor p2)
{
  // get the edge property from the two ports and make sure that they connect to the same edge.
  typedef nested_graph<NESTED_ARGS> NG;

  // remove edge from ports
  port_property_accessor<NG::PortGraph,port_edge_tag>::type edges
    = get_port_property(g, port_edge_tag());

  port_property_accessor<NG::PortGraph,port_vertex_tag>::type vertices
    = get_port_property(g, port_vertex_tag());

  port_property_accessor<NG::PortGraph,port_vgraph_tag>::type graphs
    = get_port_property(g, port_vgraph_tag());

  port_property_accessor<NG::PortGraph,port_egraph_tag>::type egraphs
    = get_port_property(g, port_egraph_tag());

  NG::edge_descriptor e1;
  e1 = edges[p1];

  assert(edges[p1] == edges[p2]);

  edges[p1] = NG::edge_descriptor();
  edges[p2] = NG::edge_descriptor();


  // find the graph that the edge belongs to.
  NG::vertex_descriptor v1 = vertices[p1];
  NG::vertex_descriptor v2 = vertices[p2];

  NG *g1 = egraphs[p1];
  NG *g2 = egraphs[p2];
  egraphs[p1] = NULL;
  egraphs[p2] = NULL;

  assert(g1 == g2);

  // now remove the actual edge and we're done
  NG::BaseGraph& base = *((NG::BaseGraph*)g1);
  remove_edge(base, e1);

}

// removes a port from a vertex
template <NESTED_TARGS> 
void remove_port(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS>::port_descriptor p)
{
  typedef nested_graph<NESTED_ARGS> NG;

  // get the property accessors for the ports.
  port_property_accessor<NG::PortGraph,port_edge_tag>::type edges
    = get_port_property(g, port_edge_tag());

  port_property_accessor<NG::PortGraph,port_vertex_tag>::type vertices
    = get_port_property(g, port_vertex_tag());

  port_property_accessor<NG::PortGraph,port_vgraph_tag>::type graphs
    = get_port_property(g, port_vgraph_tag());

  port_property_accessor<NG::PortGraph,port_egraph_tag>::type egraphs
    = get_port_property(g, port_egraph_tag());

  // remove any edges associated with this port.
  if (egraphs[p] != NULL) {
    remove_edge(*egraphs[p], edges[p]);
  }

  // delete the port from the vertex PortList.
  vertex_property_accessor<NG, vert_portlist_tag>::type plists
    = get_vertex_property(*graphs[p], vert_portlist_tag());

  NG::PortList &plist = plists[vertices[p]];
  
  // find it and remove it
  for (NG::PortList::iterator i=plist.begin(); i!=plist.end(); ++i) {
    if (*i == p) {
      plist.erase(i);
      break;
    }
  }

  // remove the port from the port graph
  remove_vertex(*g.portList, p);
}

// removes the vertex, should delete any subgraphs, plus should remove any edges associated with the vertex/child vertices.
template <NESTED_TARGS> 
void remove_vertex(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS>::vertex_descriptor v)
{
  typedef nested_graph<NESTED_ARGS> NG;

  vertex_property_accessor<NG, vert_subgraph_tag>::type graphs
    = get_vertex_property(g, vert_subgraph_tag());

  // if the vertex is a composite one, then delete the subgraph
  if (graphs[v] != NULL) {
    delete graphs[v];
    graphs[v] = NULL;
  } else {
    // otherwise delete all the ports from this vertex
    NG::vport_iterator i, iend;
    boost::tie(i, iend) = ports(g, v);
    while (i != iend) {
      remove_port(g, *i);
      boost::tie(i, iend) = ports(g, v);
    }
  }

  // now remove the vertex
  NG::BaseGraph& base = *((NG::BaseGraph*)&g);
  remove_vertex(base, v);
}

template <NESTED_TARGS> 
nested_graph<NESTED_ARGS>::PortPair ports(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS>::edge_descriptor e)
{
  typedef nested_graph<NESTED_ARGS> NG;
  edge_property_accessor<NG, edge_portpair_tag>::type ports = get_edge_property(g, edge_portpair_tag());

  return ports[e];
}

template <NESTED_TARGS> 
pair<nested_graph<NESTED_ARGS>::port_iterator, nested_graph<NESTED_ARGS>::port_iterator> ports(nested_graph<NESTED_ARGS> &g)
{
  return vertices(*g.portList);
}

template <NESTED_TARGS> 
pair<nested_graph<NESTED_ARGS>::vport_iterator, nested_graph<NESTED_ARGS>::vport_iterator> 
ports(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS>::vertex_descriptor v)
{
  typedef nested_graph<NESTED_ARGS> NG;
  vertex_property_accessor<NG, vert_portlist_tag>::type ports = get_vertex_property(g, vert_portlist_tag());

  return pair<NG::vport_iterator, NG::vport_iterator>(ports[v].begin(), ports[v].end());
  //return make_pair(ports[v].begin(), ports[v].end());
}

template <NESTED_TARGS> 
  nested_graph<NESTED_ARGS>::port_descriptor
  connected_port(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS>::port_descriptor p)
{
  typedef nested_graph<NESTED_ARGS> NG;
  port_property_accessor<NG::PortGraph, port_edge_tag>::type edges = get_port_property(g, port_edge_tag());
  port_property_accessor<NG::PortGraph, port_egraph_tag>::type egraphs = get_port_property(g, port_egraph_tag());

  NG::edge_descriptor e = edges[p];
  NG *g2 = egraphs[p];

  // if there is no edge then this port connects to nothing.
  if (e == NG::edge_descriptor()) {
    return NG::port_descriptor();
  }

  // now use the edge to find the two ports.
  NG::PortPair pp = ports(*g2, e);

  // now return the one that isn't the one passed in.
  return (pp.first == p)?pp.second:pp.first;
}

/*! copies all the vertices from g1 into g and stores edge information in the PortGraph graph to be reconstructed later.
*/
template <NESTED_TARGS> 
void expand_subgraph(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS> &g1)
{
  typedef nested_graph<NESTED_ARGS> NG;

  vertex_property_accessor<NG, vert_subgraph_tag>::type graphs 
    = get_vertex_property(g1, vert_subgraph_tag());
  vertex_property_accessor<NG, vert_portlist_tag>::type child_portlist 
    = get_vertex_property(g1, vert_portlist_tag());
  vertex_property_accessor<NG, vert_portlist_tag>::type parent_portlist 
    = get_vertex_property(g, vert_portlist_tag());
  port_property_accessor<NG::PortGraph, port_edge_tag>::type portedges 
    = get_port_property(g, port_edge_tag());
  port_property_accessor<NG::PortGraph, port_egraph_tag>::type portedgegraphs 
    = get_port_property(g, port_egraph_tag());
  port_property_accessor<NG::PortGraph, port_vertex_tag>::type portvertices 
    = get_port_property(g, port_vertex_tag());
  port_property_accessor<NG::PortGraph, port_vgraph_tag>::type portvertexgraphs 
    = get_port_property(g, port_vgraph_tag());

  // loop through all the vertices in the graph
  NG::vertex_iterator i, iend;
  for (boost::tie(i, iend) = vertices(g1); i != iend; ++i)
  {
    if (is_composite_vertex(g1, *i)) {
      expand_subgraph(g, *graphs[*i]);
    } else {
      // copy the vertex to the new graph (g)
      typedef detail::dynamic_vertex<NG::FinalVertexPlugin>* Vertex;
      VertexPlugin plug = static_cast<Vertex>(*i)->m_plugin;
      NG::vertex_descriptor v = add_vertex(g, plug);
      
      // loop through all the ports for this vertex
      NG::vport_iterator pi, piend;
      for (boost::tie(pi, piend) = ports(g1, *i); pi != piend; ++pi)
      {
        // find the connecting ports.
        NG::edge_descriptor e = portedges[*pi];
        NG *edgegraph  = portedgegraphs[*pi];
        if (edgegraph != NULL) {
          edge_property_accessor<NG, edge_portpair_tag>::type portpairs 
            = get_edge_property(*edgegraph, edge_portpair_tag());

          NG::PortPair pp = portpairs[e];

          // the edge should contain the current port
          assert(pp.first == *pi || pp.second == *pi);

          // add an edge to the PortGraph
          NG::PortGraph::adjacency_iterator a, aend;
          boost::tie(a, aend) = adj(pp.first, *g.portList);
          // if there isn't already an edge for this port pair
          if (a == aend) {
            typedef detail::dynamic_edge<NG::FinalEdgePlugin>* Edge;
            detail::dynamic_edge_base *deb = e;
            EdgePlugin plug = static_cast<Edge>(deb)->m_plugin;
            add_edge(*g.portList, pp.first, pp.second, plug);
          }

          // remove the edge from the base graph
          remove_edge(*(NG::BaseGraph*)edgegraph, e);

          portedges[pp.first] = NG::edge_descriptor();
          portedgegraphs[pp.first] = NULL;
          portedges[pp.second] = NG::edge_descriptor();
          portedgegraphs[pp.second] = NULL;

          // null out the other end as well

        }

        // now change the info in the port to point to this vertex
        portvertices[*pi] = v;
        portvertexgraphs[*pi] = &g;
      }
      parent_portlist[v] = child_portlist[*i];
      // remove the vertex
      child_portlist[*i].clear();
    }
  }
}

/*! expands a nested graph into a flat graph by expanding each of the composite subgraphs.
*/
template <NESTED_TARGS> 
void expand_graph(nested_graph<NESTED_ARGS> &g)
{
  typedef nested_graph<NESTED_ARGS> NG;

  vertex_property_accessor<NG, vert_subgraph_tag>::type graphs = get_vertex_property(g, vert_subgraph_tag());

  // loop through all the vertices in the graph
  NG::vertex_iterator i, iend;
  vector<NG::vertex_descriptor> composites;
  for (boost::tie(i, iend) = vertices(g); i != iend; ++i)
  {
    if (is_composite_vertex(g, *i)) {
      composites.push_back(*i);
    }
  }

  for (int j=0; j<composites.size(); ++j) {
    expand_subgraph(g, *graphs[composites[j]]);
    remove_vertex(g, composites[j]);
  }

  // the above code should have added a bunch of edges to the PortGraph. we can use this to add edges to the 
  // new expanded graph.
  NG::PortGraph::edge_iterator ei, eiend;
  for (boost::tie(ei, eiend) = edges(*g.portList); ei != eiend; ++ei)
  {
    // add a similar edge to the main graph.
    NG::port_descriptor p1 = source(*ei, *g.portList);
    NG::port_descriptor p2 = target(*ei, *g.portList);

    typedef detail::dynamic_edge<NG::FinalEdgePlugin>* Edge;
    detail::dynamic_edge_base *deb = *ei;
    EdgePlugin plug = static_cast<Edge>(deb)->m_plugin;
    add_edge_from_ports(g, p1, p2, plug);
  }

  // remove all the edges from PortGraph
  NG::PortGraph::vertex_iterator vi, viend;
  for (boost::tie(vi, viend) = vertices(*g.portList); vi != viend; ++vi)
  {
    clear_vertex(*g.portList, *vi);
  }
}


template <NESTED_TARGS> 
bool validate_graph(nested_graph<NESTED_ARGS> &g, string &problem)
{
  typedef nested_graph<NESTED_ARGS> NG;

  vertex_property_accessor<NG, vert_subgraph_tag>::type graphs 
    = get_vertex_property(g, vert_subgraph_tag());
  port_property_accessor<NG::PortGraph, port_edge_tag>::type portedges 
    = get_port_property(g, port_edge_tag());
  port_property_accessor<NG::PortGraph, port_egraph_tag>::type portegraphs 
    = get_port_property(g, port_egraph_tag());
  port_property_accessor<NG::PortGraph, port_vertex_tag>::type portvertices 
    = get_port_property(g, port_vertex_tag());
  port_property_accessor<NG::PortGraph, port_vgraph_tag>::type portvgraphs 
    = get_port_property(g, port_vgraph_tag());
  
  NG::port_iterator i, iend;
  for (boost::tie(i, iend) = ports(g); i != iend; ++i) {
    // get the edge that matches this port
    NG::edge_descriptor e = portedges[*i];
    NG                  *eg = portegraphs[*i];

    // make sure one of the ports on the edge matches this port
    if (eg != NULL) {
      edge_property_accessor<NG, edge_portpair_tag>::type edge_portpairs 
        = get_edge_property(*eg, edge_portpair_tag());

      NG::PortPair pp = edge_portpairs[e];

      if (pp.first != *i && pp.second != *i){
        problem = "Port points to edge which doesn't point back to port";
        return false;
      }
    }
    
    // check that the vertex the port belongs to has the port in it's list
    NG::vertex_descriptor v = portvertices[*i];
    NG                  *vg = portvgraphs[*i];
    
    vertex_property_accessor<NG, vert_portlist_tag>::type vert_portlist 
      = get_vertex_property(*vg, vert_portlist_tag());

    NG::vport_iterator vpi, vpiend;
    bool bFound = false;
    for (boost::tie(vpi, vpiend) = ports(*vg, v); vpi != vpiend && !bFound; ++vpi) {
        bFound = (*vpi == *i);
    }

    if (!bFound) {
      problem = "port's vertex does not contain reference to the port.";
      return false;
    }
  }

  return true;
}

template <NESTED_TARGS> 
void clear_port_lists(nested_graph<NESTED_ARGS> &g)
{
  typedef nested_graph<NESTED_ARGS> NG;

  vertex_property_accessor<NG, vert_subgraph_tag>::type graphs 
    = get_vertex_property(g, vert_subgraph_tag());
  vertex_property_accessor<NG, vert_portlist_tag>::type portlists
    = get_vertex_property(g, vert_portlist_tag());

  // loop through all the vertices in the graph
  NG::vertex_iterator i, iend;
  for (boost::tie(i, iend) = vertices(g); i != iend; ++i)
  {
    if (is_composite_vertex(g, *i)) {
      clear_port_lists(*graphs[*i]);
    } else {
      portlists[*i].clear();
    }
  }
}

template <NESTED_TARGS> 
void update_subgraphs(nested_graph<NESTED_ARGS> &g)
{
  typedef nested_graph<NESTED_ARGS> NG;

  // get the property accessor for graph tag
  vertex_property_accessor<NG, vert_subgraph_tag>::type subGraph
    = get_vertex_property(g, vert_subgraph_tag());

  // loop through all the vertices in the graph
  NG::vertex_iterator i, iend;
  for (boost::tie(i, iend) = vertices(g); i != iend; ++i)
  {
    if (is_composite_vertex(g, *i)) {
      NG *subgraph = subGraph[*i];
      subgraph->nestLevel = g.nestLevel + 1;
      subgraph->portList = g.portList;
      update_subgraphs(*subgraph);
    }
  }
}
  
  
template <NESTED_TARGS> 
void insert_subgraph(nested_graph<NESTED_ARGS> &g, nested_graph<NESTED_ARGS>::vertex_descriptor v, nested_graph<NESTED_ARGS> *subgraph)
{
  typedef nested_graph<NESTED_ARGS> NG;

  // get the property accessor for graph tag
  vertex_property_accessor<NG, vert_subgraph_tag>::type subGraph
    = get_vertex_property(g, vert_subgraph_tag());

  // firstly delete the existing subgraph of the vertex.
  assert(subGraph[v] != NULL);
  subgraph->parentGraph = subGraph[v]->parentGraph; 
  subgraph->parentVertex = subGraph[v]->parentVertex; 
  delete subGraph[v];
  subGraph[v] = subgraph;

  // change the nesting level
  subgraph->nestLevel = g.nestLevel + 1;
  // add the new graph, and copy all the ports to the current graph
  NG::PortGraph *oldPG = subgraph->portList;
  NG::PortGraph *newPG = g.portList;

  port_property_accessor<NG::PortGraph, port_edge_tag>::type portedges 
    = get_port_property(g, port_edge_tag());
  port_property_accessor<NG::PortGraph, port_egraph_tag>::type portegraphs 
    = get_port_property(g, port_egraph_tag());
  port_property_accessor<NG::PortGraph, port_vertex_tag>::type portvertices 
    = get_port_property(g, port_vertex_tag());
  port_property_accessor<NG::PortGraph, port_vgraph_tag>::type portvgraphs 
    = get_port_property(g, port_vgraph_tag());

  // remove all the port's from all the vertixes in this graph and its subgraphs.
  clear_port_lists(*subgraph);

  NG::port_iterator i, iend;
  for (boost::tie(i, iend) = ports(*subgraph); i != iend; ++i) {
    NG::port_descriptor oldP = *i;

    // copy the new port to the new port graph
    typedef detail::dynamic_vertex<NG::FinalPortPlugin>* Port;
    NG::FinalPortPlugin plug = static_cast<Port>(*i)->m_plugin;
    NG::port_descriptor newP = add_vertex(*newPG, plug);

    // now change the port descriptor in the vertex port list and the edge properties
    NG::vertex_descriptor vert = portvertices[newP];
    NG *                  vertGraph = portvgraphs[newP];

    // add the new port to the 
    vertex_property_accessor<NG, vert_portlist_tag>::type portList
      = get_vertex_property(*vertGraph, vert_portlist_tag());

    portList[vert].push_back(newP);

    NG::edge_descriptor edge = portedges[newP];
    NG *           edgeGraph = portegraphs[newP];

    // now change the edge to point the new port.
    edge_property_accessor<NG, edge_portpair_tag>::type portPairs
      = get_edge_property(*edgeGraph, edge_portpair_tag());

    NG::PortPair pp = portPairs[edge];
    NG::PortPair newPP;
    
    NG::port_descriptor emptyPort = NG::port_descriptor();
    if (pp.first == emptyPort) {
      newPP = make_pair(newP, pp.second);     // first port is empty so put the new one here
    } else if (pp.second == emptyPort) {
      newPP = make_pair(pp.first, newP);      // second port is empty so put it here
    } else if (pp.first == oldP) {
      newPP = make_pair(newP, emptyPort);     // first port matches so put new one in and wipe the other port
    } else {
      newPP = make_pair(emptyPort , newP);     // second port should match so put new one here and wipe the firt one.
    }
    portPairs[edge] = newPP;
  }
  
  subgraph->portList = newPG;
  delete oldPG;

  update_subgraphs(*subgraph);

}


/*! returns a ports parent vertex and the graph it belongs to */
template <NESTED_TARGS> 
  pair<nested_graph<NESTED_ARGS>::port_descriptor, nested_graph<NESTED_ARGS>*> 
                                                                          get_parent_vertex(nested_graph<NESTED_ARGS> &g, 
                                                                                   nested_graph<NESTED_ARGS>::port_descriptor p)
{
  typedef nested_graph<NESTED_ARGS> NG;

  // get all the 
  vertex_property_accessor<NG::PortGraph, port_vertex_tag>::type vlist
    = get_vertex_property(*g.portList, port_vertex_tag());
  vertex_property_accessor<NG::PortGraph, port_vgraph_tag>::type glist
    = get_vertex_property(*g.portList, port_vgraph_tag());

  return make_pair(vlist[p], glist[p]);
}
